@@ -1,4 +1,4 @@
-# Copyright (c) 2014, 2020, Oracle and/or its affiliates.
+# Copyright (c) 2014, 2021, Oracle and/or its affiliates.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2.0, as
@@ -29,9 +29,11 @@
 """Module gathering all abstract base classes"""
 
 from abc import ABCMeta, abstractmethod, abstractproperty
+from decimal import Decimal
+from time import sleep
+from datetime import date, datetime, time, timedelta
 import os
 import re
-import time
 import weakref
 TLS_V1_3_SUPPORTED = False
 try:
@@ -42,12 +44,12 @@ except:
     # If import fails, we don't have SSL support.
     pass
 
-from .catch23 import make_abc, BYTE_TYPES, STRING_TYPES
 from .conversion import MySQLConverterBase
 from .constants import (ClientFlag, CharacterSet, CONN_ATTRS_DN,
                         DEFAULT_CONFIGURATION, OPENSSL_CS_NAMES,
                         TLS_CIPHER_SUITES, TLS_VERSIONS)
 from .optionfiles import MySQLOptionsParser
+from .utils import make_abc
 from . import errors
 
 NAMED_TUPLE_CACHE = weakref.WeakValueDictionary()
@@ -62,6 +64,15 @@ TLS_VERSION_ERROR = ("The given tls_version: '{}' is not recognized as a valid "
 TLS_VER_NO_SUPPORTED = ("No supported TLS protocol version found in the "
                         "'tls-versions' list '{}'. ")
 
+KRB_SERVICE_PINCIPAL_ERROR = (
+    'Option "krb_service_principal" {error}, must be a string in the form '
+    '"primary/instance@realm" e.g "ldap/ldapauth@MYSQL.COM" where "@realm" '
+    'is optional and if it is not given will be assumed to belong to the '
+    'default realm, as configured in the krb5.conf file.')
+
+MYSQL_PY_TYPES = (
+    (int, str, bytes, Decimal, float, datetime, date, timedelta, time,))
+
 
 @make_abc(ABCMeta)
 class MySQLConnectionAbstract(object):
@@ -81,6 +92,9 @@ class MySQLConnectionAbstract(object):
 
         self._user = ''
         self._password = ''
+        self._password1 = ''
+        self._password2 = ''
+        self._password3 = ''
         self._database = ''
         self._host = '127.0.0.1'
         self._port = 3306
@@ -90,6 +104,7 @@ class MySQLConnectionAbstract(object):
         self._ssl = {}
         self._ssl_disabled = DEFAULT_CONFIGURATION["ssl_disabled"]
         self._force_ipv6 = False
+        self._oci_config_file = None
 
         self._use_unicode = True
         self._get_warnings = False
@@ -105,12 +120,14 @@ class MySQLConnectionAbstract(object):
             DEFAULT_CONFIGURATION["allow_local_infile_in_path"])
 
         self._prepared_statements = None
+        self._query_attrs = []
 
         self._ssl_active = False
         self._auth_plugin = None
         self._pool_config_version = None
         self.converter = None
         self._converter_class = None
+        self._converter_str_fallback = False
         self._compress = False
 
         self._consume_results = False
@@ -188,7 +205,7 @@ class MySQLConnectionAbstract(object):
         tls_ciphersuites = []
         tls_cs = self._ssl["tls_ciphersuites"]
 
-        if isinstance(tls_cs, STRING_TYPES):
+        if isinstance(tls_cs, str):
             if not (tls_cs.startswith("[") and
                     tls_cs.endswith("]")):
                 raise AttributeError("tls_ciphersuites must be a list, "
@@ -270,7 +287,7 @@ class MySQLConnectionAbstract(object):
         tls_versions = []
         tls_version = self._ssl["tls_versions"]
 
-        if isinstance(tls_version, STRING_TYPES):
+        if isinstance(tls_version, str):
             if not (tls_version.startswith("[") and tls_version.endswith("]")):
                 raise AttributeError("tls_versions must be a list, found: '{}'"
                                      "".format(tls_version))
@@ -579,7 +596,7 @@ class MySQLConnectionAbstract(object):
                 if attr_name in CONN_ATTRS_DN:
                     continue
                 # Validate name type
-                if not isinstance(attr_name, STRING_TYPES):
+                if not isinstance(attr_name, str):
                     raise errors.InterfaceError(
                         "Attribute name should be a string, found: '{}' in '{}'"
                         "".format(attr_name, self._conn_attrs))
@@ -595,7 +612,7 @@ class MySQLConnectionAbstract(object):
                         "'_', found: '{}'".format(attr_name))
                 # Validate value type
                 attr_value = self._conn_attrs[attr_name]
-                if not isinstance(attr_value, STRING_TYPES):
+                if not isinstance(attr_value, str):
                     raise errors.InterfaceError(
                         "Attribute '{}' value: '{}' must be a string type."
                         "".format(attr_name, attr_value))
@@ -607,6 +624,18 @@ class MySQLConnectionAbstract(object):
 
         if self._client_flags & ClientFlag.CONNECT_ARGS:
             self._add_default_conn_attrs()
+        if "krb_service_principal" in config and \
+            config["krb_service_principal"] is not None:
+            self._krb_service_principal = config["krb_service_principal"]
+            if not isinstance(self._krb_service_principal, str):
+                raise errors.InterfaceError(KRB_SERVICE_PINCIPAL_ERROR.format(
+                    error="is not a string"))
+            if self._krb_service_principal == "":
+                raise errors.InterfaceError(KRB_SERVICE_PINCIPAL_ERROR.format(
+                    error="can not be an empty string"))
+            if "/" not in self._krb_service_principal:
+                raise errors.InterfaceError(KRB_SERVICE_PINCIPAL_ERROR.format(
+                    error="is incorrectly formatted"))
 
     def _add_default_conn_attrs(self):
         """Add the default connection attributes."""
@@ -623,7 +652,7 @@ class MySQLConnectionAbstract(object):
 
         Returns tuple
         """
-        if isinstance(server_version, BYTE_TYPES):
+        if isinstance(server_version, (bytearray, bytes)):
             server_version = server_version.decode()
 
         # pylint: disable=W1401
@@ -1015,7 +1044,7 @@ class MySQLConnectionAbstract(object):
                           "attempt(s): {1}".format(attempts, str(err))
                     raise errors.InterfaceError(msg)
             if delay > 0:
-                time.sleep(delay)
+                sleep(delay)
 
     @abstractmethod
     def is_connected(self):
@@ -1149,6 +1178,7 @@ class MySQLConnectionAbstract(object):
             charset_name = CharacterSet.get_info(self._charset_id)[0]
             self._converter_class = convclass
             self.converter = convclass(charset_name, self._use_unicode)
+            self.converter.str_fallback = self._converter_str_fallback
         else:
             raise TypeError("Converter class should be a subclass "
                             "of conversion.MySQLConverterBase.")
@@ -1213,7 +1243,7 @@ class MySQLConnectionAbstract(object):
         raise NotImplementedError
 
     def cmd_change_user(self, username='', password='', database='',
-                        charset=45):
+                        charset=45, password1='', password2='', password3=''):
         """Change the current logged in user"""
         raise NotImplementedError
 
@@ -1435,3 +1465,30 @@ class MySQLCursorAbstract(object):
     def fetchwarnings(self):
         """Returns Warnings."""
         return self._warnings
+
+    def get_attributes(self):
+        """Get the added query attributes so far."""
+        if hasattr(self, "_cnx"):
+            return self._cnx._query_attrs
+        elif hasattr(self, "_connection"):
+            return self._connection._query_attrs
+
+    def add_attribute(self, name, value):
+        """Add a query attribute and his value."""
+        if not isinstance(name, str):
+            raise errors.ProgrammingError(
+                "Parameter `name` must be a string type.")
+        if value is not None and not isinstance(value, MYSQL_PY_TYPES):
+            raise errors.ProgrammingError(
+                f"Object {value} cannot be converted to a MySQL type.")
+        if hasattr(self, "_cnx"):
+            self._cnx._query_attrs.append((name, value))
+        elif hasattr(self, "_connection"):
+            self._connection._query_attrs.append((name, value))
+
+    def clear_attributes(self):
+        """Remove all the query attributes."""
+        if hasattr(self, "_cnx"):
+            self._cnx._query_attrs = []
+        elif hasattr(self, "_connection"):
+            self._connection._query_attrs = []
