@@ -1,4 +1,4 @@
-# Copyright (c) 2014, 2020, Oracle and/or its affiliates.
+# Copyright (c) 2014, 2021, Oracle and/or its affiliates.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2.0, as
@@ -37,7 +37,6 @@ import socket
 import sysconfig
 
 from . import errors, version
-from .catch23 import INT_TYPES
 from .constants import (
     CharacterSet, FieldFlag, ServerFlag, ShutdownType, ClientFlag
 )
@@ -77,7 +76,7 @@ class CMySQLConnection(MySQLConnectionAbstract):
         self._columns = []
         self._plugin_dir = os.path.join(
             os.path.dirname(os.path.abspath(_mysql_connector.__file__)),
-            "mysql", "vendor"
+            "mysql", "vendor", "plugin"
         )
         self.converter = None
         super(CMySQLConnection, self).__init__(**kwargs)
@@ -191,6 +190,9 @@ class CMySQLConnection(MySQLConnectionAbstract):
             'host': self._host,
             'user': self._user,
             'password': self._password,
+            'password1': self._password1,
+            'password2': self._password2,
+            'password3': self._password3,
             'database': self._database,
             'port': self._port,
             'client_flags': self._client_flags,
@@ -199,7 +201,8 @@ class CMySQLConnection(MySQLConnectionAbstract):
             'ssl_disabled': True,
             "conn_attrs": self._conn_attrs,
             "local_infile": self._allow_local_infile,
-            "load_data_local_dir": self._allow_local_infile_in_path
+            "load_data_local_dir": self._allow_local_infile_in_path,
+            "oci_config_file": self._oci_config_file,
         }
 
         tls_versions = self._ssl.get('tls_versions')
@@ -231,6 +234,9 @@ class CMySQLConnection(MySQLConnectionAbstract):
 
         try:
             self._cmysql.connect(**cnx_kwargs)
+            self._cmysql.converter_str_fallback = self._converter_str_fallback
+            if self.converter:
+                self.converter.str_fallback = self._converter_str_fallback
         except MySQLInterfaceError as exc:
             raise errors.get_mysql_exception(msg=exc.msg, errno=exc.errno,
                                              sqlstate=exc.sqlstate)
@@ -246,9 +252,12 @@ class CMySQLConnection(MySQLConnectionAbstract):
             except MySQLInterfaceError as exc:
                 raise errors.get_mysql_exception(msg=exc.msg, errno=exc.errno,
                                                  sqlstate=exc.sqlstate)
-            self._cmysql = None
     disconnect = close
 
+    def is_closed(self):
+        """Return True if the connection to MySQL Server is closed."""
+        return not self._cmysql.connected()
+
     def is_connected(self):
         """Reports whether the connection to MySQL Server is available"""
         if self._cmysql:
@@ -428,6 +437,7 @@ class CMySQLConnection(MySQLConnectionAbstract):
                 None,
                 None,
                 None,
+                None,
                 ~int(col[9]) & FieldFlag.NOT_NULL,
                 int(col[9])
             ))
@@ -459,7 +469,9 @@ class CMySQLConnection(MySQLConnectionAbstract):
             raise errors.OperationalError("MySQL Connection not available")
 
         try:
-            return self._cmysql.stmt_prepare(statement)
+            stmt = self._cmysql.stmt_prepare(statement)
+            stmt.converter_str_fallback = self._converter_str_fallback
+            return stmt
         except MySQLInterfaceError as err:
             raise errors.InterfaceError(str(err))
 
@@ -503,7 +515,8 @@ class CMySQLConnection(MySQLConnectionAbstract):
                 query = query.encode('utf-8')
             self._cmysql.query(query,
                                raw=raw, buffered=buffered,
-                               raw_as_string=raw_as_string)
+                               raw_as_string=raw_as_string,
+                               query_attrs=self._query_attrs)
         except MySQLInterfaceError as exc:
             raise errors.get_mysql_exception(exc.errno, msg=exc.msg,
                                              sqlstate=exc.sqlstate)
@@ -640,11 +653,28 @@ class CMySQLConnection(MySQLConnectionAbstract):
         Returns dict.
         """
         if isinstance(params, (list, tuple)):
-            result = self._cmysql.convert_to_mysql(*params)
+            if self.converter:
+                result = [
+                    self.converter.quote(
+                        self.converter.escape(
+                            self.converter.to_mysql(value)
+                        )
+                    ) for value in params
+                ]
+            else:
+                result = self._cmysql.convert_to_mysql(*params)
         elif isinstance(params, dict):
             result = {}
-            for key, value in params.items():
-                result[key] = self._cmysql.convert_to_mysql(value)[0]
+            if self.converter:
+                for key, value in params.items():
+                    result[key] = self.converter.quote(
+                        self.converter.escape(
+                            self.converter.to_mysql(value)
+                        )
+                    )
+            else:
+                for key, value in params.items():
+                    result[key] = self._cmysql.convert_to_mysql(value)[0]
         else:
             raise ValueError("Could not process parameters")
 
@@ -658,10 +688,19 @@ class CMySQLConnection(MySQLConnectionAbstract):
         self._cmysql.consume_result()
 
     def cmd_change_user(self, username='', password='', database='',
-                        charset=45):
+                        charset=45, password1='', password2='', password3='',
+                        oci_config_file=None):
         """Change the current logged in user"""
         try:
-            self._cmysql.change_user(username, password, database)
+            self._cmysql.change_user(
+                username,
+                password,
+                database,
+                password1,
+                password2,
+                password3,
+                oci_config_file)
+
         except MySQLInterfaceError as exc:
             raise errors.get_mysql_exception(msg=exc.msg, errno=exc.errno,
                                              sqlstate=exc.sqlstate)
@@ -732,7 +771,7 @@ class CMySQLConnection(MySQLConnectionAbstract):
 
     def cmd_process_kill(self, mysql_pid):
         """Kill a MySQL process"""
-        if not isinstance(mysql_pid, INT_TYPES):
+        if not isinstance(mysql_pid, int):
             raise ValueError("MySQL PID must be int")
         self.info_query("KILL {0}".format(mysql_pid))
 
@@ -777,7 +816,9 @@ class CMySQLConnection(MySQLConnectionAbstract):
                     "version 5.7.2 or earlier.")
             else:
                 self.cmd_change_user(self._user, self._password,
-                                     self._database, self._charset_id)
+                                     self._database, self._charset_id,
+                                     self._password1, self._password2,
+                                     self._password3, self._oci_config_file)
 
         if user_variables or session_variables:
             cur = self.cursor()
