@@ -28,8 +28,7 @@
 
 """Expression Parser."""
 
-from .compat import STRING_TYPES, BYTE_TYPES, UNICODE_TYPES
-from .helpers import get_item_or_attr
+from .helpers import BYTE_TYPES, get_item_or_attr
 from .dbdoc import DbDoc
 from .protobuf import Message, mysqlxpb_enum
 
@@ -309,7 +308,7 @@ def build_expr(value):
 
 
 def build_scalar(value):
-    if isinstance(value, STRING_TYPES):
+    if isinstance(value, str):
         return build_string_scalar(value)
     elif isinstance(value, BYTE_TYPES):
         return build_bytes_scalar(value)
@@ -331,7 +330,7 @@ def build_object(obj):
     msg = Message("Mysqlx.Expr.Object")
     for key, value in obj.items():
         pair = Message("Mysqlx.Expr.Object.ObjectField")
-        pair["key"] = key.encode() if isinstance(key, UNICODE_TYPES) else key
+        pair["key"] = key.encode() if isinstance(key, str) else key
         pair["value"] = build_expr(value).get_message()
         msg["fld"].extend([pair.get_message()])
     return msg
@@ -369,7 +368,7 @@ def build_unsigned_int_scalar(value):
     return msg
 
 def build_string_scalar(value):
-    if isinstance(value, STRING_TYPES):
+    if isinstance(value, str):
         value = bytes(bytearray(value, "utf-8"))
     msg = Message("Mysqlx.Datatypes.Scalar")
     msg["type"] = mysqlxpb_enum("Mysqlx.Datatypes.Scalar.Type.V_STRING")
@@ -413,9 +412,10 @@ def escape_literal(string):
 
 
 class ExprParser(object):
-    def __init__(self, string, allow_relational=False):
+    def __init__(self, string, allow_relational=True):
         self.string = string
         self.tokens = []
+        self.path_name_queue = []
         self.pos = 0
         self._allow_relational_columns = allow_relational
         self.placeholder_name_to_position = {}
@@ -432,7 +432,7 @@ class ExprParser(object):
         Removes the keywords "SELECT" and "WHERE" that does not form part of
         the expression itself.
         """
-        if not isinstance(self.string, STRING_TYPES):
+        if not isinstance(self.string, str):
             self.string = repr(self.string)
         self.string = self.string.strip(" ")
         if len(self.string) > 1 and self.string[-1] == ';':
@@ -624,8 +624,9 @@ class ExprParser(object):
                              "tokens left".format(token_type, self.pos))
         if self.tokens[self.pos].token_type != token_type:
             raise ValueError("Expected token type {0} at pos {1} but found "
-                             "type {2}".format(token_type, self.pos,
-                                               self.tokens[self.pos]))
+                             "type {2}, on tokens {3}"
+                             "".format(token_type, self.pos,
+                                       self.tokens[self.pos], self.tokens))
 
     def cur_token_type_is(self, token_type):
         return self.pos_token_type_is(self.pos, token_type)
@@ -656,13 +657,24 @@ class ExprParser(object):
         list and return a list of Expr objects.
         """
         exprs = []
+        path_name_added = False
         self.consume_token(TokenType.LPAREN)
         if not self.cur_token_type_is(TokenType.RPAREN):
-            exprs.append(self._expr().get_message())
+            msg_expr = self._expr().get_message()
+            if hasattr(msg_expr, "identifier") and msg_expr.identifier.name:
+                self.path_name_queue.insert(0, msg_expr.identifier.name)
+                path_name_added = True
+            elif not hasattr(msg_expr, "identifier") and \
+               "identifier" in msg_expr and "name" in msg_expr["identifier"]:
+                self.path_name_queue.insert(0, msg_expr["identifier"]["name"])
+                path_name_added = True
+            exprs.append(msg_expr)
             while self.cur_token_type_is(TokenType.COMMA):
                 self.pos += 1
                 exprs.append(self._expr().get_message())
         self.consume_token(TokenType.RPAREN)
+        if path_name_added:
+            self.path_name_queue.pop()
         return exprs
 
     def identifier(self):
@@ -740,6 +752,8 @@ class ExprParser(object):
             doc_path_item["value"] = self.consume_token(TokenType.IDENT)
             col_id["document_path"].extend([doc_path_item.get_message()])
         col_id["document_path"].extend(self.document_path())
+        if self.path_name_queue:
+            col_id["name"] = self.path_name_queue[0]
         expr = Message("Mysqlx.Expr.Expr")
         expr["type"] = mysqlxpb_enum("Mysqlx.Expr.Expr.Type.IDENT")
         expr["identifier"] = col_id
