@@ -1,4 +1,4 @@
-# Copyright (c) 2016, 2018, Oracle and/or its affiliates. All rights reserved.
+# Copyright (c) 2016, 2020, Oracle and/or its affiliates.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2.0, as
@@ -34,11 +34,10 @@ import warnings
 
 from .errors import ProgrammingError, NotSupportedError
 from .expr import ExprParser
-from .compat import INT_TYPES, STRING_TYPES
 from .constants import LockContention
 from .dbdoc import DbDoc
 from .helpers import deprecated
-from .result import SqlResult, Result
+from .result import Result
 from .protobuf import mysqlxpb_enum
 
 ERR_INVALID_INDEX_NAME = 'The given index name "{}" is not valid'
@@ -119,7 +118,7 @@ def parse_table_name(default_schema, table_name, sql_mode=""):
     quote = '"' if "ANSI_QUOTES" in sql_mode else "`"
     delimiter = ".{0}".format(quote) if quote in table_name else "."
     temp = table_name.split(delimiter, 1)
-    return (default_schema if len(temp) is 1 else temp[0].strip(quote),
+    return (default_schema if len(temp) == 1 else temp[0].strip(quote),
             temp[-1].strip(quote),)
 
 
@@ -135,6 +134,11 @@ class Statement(object):
         self._target = target
         self._doc_based = doc_based
         self._connection = target.get_connection() if target else None
+        self._stmt_id = None
+        self._exec_counter = 0
+        self._changed = True
+        self._prepared = False
+        self._deallocate_prepare_execute = False
 
     @property
     def target(self):
@@ -146,6 +150,58 @@ class Statement(object):
         """:class:`mysqlx.Schema`: The Schema object."""
         return self._target.schema
 
+    @property
+    def stmt_id(self):
+        """Returns this statement ID.
+
+        Returns:
+            int: The statement ID.
+        """
+        return self._stmt_id
+
+    @stmt_id.setter
+    def stmt_id(self, value):
+        self._stmt_id = value
+
+    @property
+    def exec_counter(self):
+        """int: The number of times this statement was executed."""
+        return self._exec_counter
+
+    @property
+    def changed(self):
+        """bool: `True` if this statement has changes."""
+        return self._changed
+
+    @changed.setter
+    def changed(self, value):
+        self._changed = value
+
+    @property
+    def prepared(self):
+        """bool: `True` if this statement has been prepared."""
+        return self._prepared
+
+    @prepared.setter
+    def prepared(self, value):
+        self._prepared = value
+
+    @property
+    def repeated(self):
+        """bool: `True` if this statement was executed more than once.
+        """
+        return self._exec_counter > 1
+
+    @property
+    def deallocate_prepare_execute(self):
+        """bool: `True` to deallocate + prepare + execute statement.
+        """
+        return self._deallocate_prepare_execute
+
+    @deallocate_prepare_execute.setter
+    def deallocate_prepare_execute(self, value):
+        self._deallocate_prepare_execute = value
+
     def is_doc_based(self):
         """Check if it is document based.
 
@@ -154,6 +210,14 @@ class Statement(object):
         """
         return self._doc_based
 
+    def increment_exec_counter(self):
+        """Increments the number of times this statement has been executed."""
+        self._exec_counter += 1
+
+    def reset_exec_counter(self):
+        """Resets the number of times this statement has been executed."""
+        self._exec_counter = 0
+
     def execute(self):
         """Execute the statement.
 
@@ -178,12 +242,12 @@ class FilterableStatement(Statement):
         super(FilterableStatement, self).__init__(target=target,
                                                   doc_based=doc_based)
         self._binding_map = {}
-        self._bindings = []
+        self._bindings = {}
         self._having = None
         self._grouping_str = ""
         self._grouping = None
         self._limit_offset = 0
-        self._limit_row_count = 0
+        self._limit_row_count = None
         self._projection_str = ""
         self._projection_expr = None
         self._sort_str = ""
@@ -204,15 +268,17 @@ class FilterableStatement(Statement):
         """Bind single object.
 
         Args:
-            obj(:class:`mysqlx.DbDoc` or str): DbDoc or JSON string object.
+            obj (:class:`mysqlx.DbDoc` or str): DbDoc or JSON string object.
 
         Raises:
             :class:`mysqlx.ProgrammingError`: If invalid JSON string to bind.
             ValueError: If JSON loaded is not a dictionary.
         """
-        if isinstance(obj, DbDoc):
-            self.bind(str(obj))
-        elif isinstance(obj, STRING_TYPES):
+        if isinstance(obj, dict):
+            self.bind(DbDoc(obj).as_str())
+        elif isinstance(obj, DbDoc):
+            self.bind(obj.as_str())
+        elif isinstance(obj, str):
             try:
                 res = json.loads(obj)
                 if not isinstance(res, dict):
@@ -237,6 +303,7 @@ class FilterableStatement(Statement):
         self._sort_str = ",".join(flexible_params(*clauses))
         self._sort_expr = ExprParser(self._sort_str,
                                      not self._doc_based).parse_order_spec()
+        self._changed = True
         return self
 
     def _set_where(self, condition):
@@ -257,6 +324,7 @@ class FilterableStatement(Statement):
         except ValueError:
             raise ProgrammingError("Invalid condition")
         self._binding_map = expr.placeholder_name_to_position
+        self._changed = True
         return self
 
     def _set_group_by(self, *fields):
@@ -270,6 +338,7 @@ class FilterableStatement(Statement):
         self._grouping_str = ",".join(fields)
         self._grouping = ExprParser(self._grouping_str,
                                     not self._doc_based).parse_expr_list()
+        self._changed = True
 
     def _set_having(self, condition):
         """Set having.
@@ -279,6 +348,7 @@ class FilterableStatement(Statement):
         """
         self.has_having = True
         self._having = ExprParser(condition, not self._doc_based).expr()
+        self._changed = True
 
     def _set_projection(self, *fields):
         """Set the projection.
@@ -295,6 +365,7 @@ class FilterableStatement(Statement):
         self._projection_expr = ExprParser(
             self._projection_str,
             not self._doc_based).parse_table_select_projection()
+        self._changed = True
         return self
 
     def get_binding_map(self):
@@ -413,10 +484,14 @@ class FilterableStatement(Statement):
         .. versionchanged:: 8.0.12
            The usage of ``offset`` was deprecated.
         """
-        if not isinstance(row_count, INT_TYPES) or row_count < 0:
+        if not isinstance(row_count, int) or row_count < 0:
             raise ValueError("The 'row_count' value must be a positive integer")
-        self.has_limit = True
+        if not self.has_limit:
+            self._changed = bool(self._exec_counter == 0)
+            self._deallocate_prepare_execute = bool(not self._exec_counter == 0)
+
         self._limit_row_count = row_count
+        self.has_limit = True
         if offset:
             self.offset(offset)
             warnings.warn("'limit(row_count, offset)' is deprecated, please "
@@ -438,13 +513,13 @@ class FilterableStatement(Statement):
 
         .. versionadded:: 8.0.12
         """
-        if not isinstance(offset, INT_TYPES) or offset < 0:
+        if not isinstance(offset, int) or offset < 0:
             raise ValueError("The 'offset' value must be a positive integer")
         self._limit_offset = offset
         return self
 
     def bind(self, *args):
-        """Binds a value to a specific placeholder.
+        """Binds value(s) to a specific placeholder(s).
 
         Args:
             *args: The name of the placeholder and the value to bind.
@@ -461,10 +536,10 @@ class FilterableStatement(Statement):
         count = len(args)
         if count == 1:
             self._bind_single(args[0])
-        elif count > 2:
-            raise ProgrammingError("Invalid number of arguments to bind")
+        elif count == 2:
+            self._bindings[args[0]] = args[1]
         else:
-            self._bindings.append({"name": args[0], "value": args[1]})
+            raise ProgrammingError("Invalid number of arguments to bind")
         return self
 
     def execute(self):
@@ -487,6 +562,50 @@ class SqlStatement(Statement):
         super(SqlStatement, self).__init__(target=None, doc_based=False)
         self._connection = connection
         self._sql = sql
+        self._binding_map = None
+        self._bindings = []
+        self.has_bindings = False
+        self.has_limit = False
+
+    @property
+    def sql(self):
+        """string: The SQL text statement."""
+        return self._sql
+
+    def get_binding_map(self):
+        """Returns the binding map dictionary.
+
+        Returns:
+            dict: The binding map dictionary.
+        """
+        return self._binding_map
+
+    def get_bindings(self):
+        """Returns the bindings list.
+
+        Returns:
+            `list`: The bindings list.
+        """
+        return self._bindings
+
+    def bind(self, *args):
+        """Binds value(s) to a specific placeholder(s).
+
+        Args:
+            *args: The value(s) to bind.
+
+        Returns:
+            mysqlx.SqlStatement: SqlStatement object.
+        """
+        if len(args) == 0:
+            raise ProgrammingError("Invalid number of arguments to bind")
+        self.has_bindings = True
+        bindings = flexible_params(*args)
+        if isinstance(bindings, (list, tuple)):
+            self._bindings = bindings
+        else:
+            self._bindings.append(bindings)
+        return self
 
     def execute(self):
         """Execute the statement.
@@ -494,8 +613,7 @@ class SqlStatement(Statement):
         Returns:
             mysqlx.SqlResult: SqlResult object.
         """
-        self._connection.send_sql(self._sql)
-        return SqlResult(self._connection)
+        return self._connection.send_sql(self)
 
 
 class WriteStatement(Statement):
@@ -628,7 +746,7 @@ class ModifyStatement(FilterableStatement):
     def __init__(self, collection, condition):
         super(ModifyStatement, self).__init__(target=collection,
                                               condition=condition)
-        self._update_ops = []
+        self._update_ops = {}
 
     def sort(self, *clauses):
         """Sets the sorting criteria.
@@ -659,9 +777,10 @@ class ModifyStatement(FilterableStatement):
         Returns:
             mysqlx.ModifyStatement: ModifyStatement object.
         """
-        self._update_ops.append(UpdateSpec(mysqlxpb_enum(
+        self._update_ops[doc_path] = UpdateSpec(mysqlxpb_enum(
             "Mysqlx.Crud.UpdateOperation.UpdateType.ITEM_SET"),
-                                           doc_path, value))
+                                                doc_path, value)
+        self._changed = True
         return self
 
     @deprecated("8.0.12")
@@ -678,25 +797,26 @@ class ModifyStatement(FilterableStatement):
 
         .. deprecated:: 8.0.12
         """
-        self._update_ops.append(UpdateSpec(mysqlxpb_enum(
+        self._update_ops[doc_path] = UpdateSpec(mysqlxpb_enum(
             "Mysqlx.Crud.UpdateOperation.UpdateType.ITEM_REPLACE"),
-                                           doc_path, value))
+                                                doc_path, value)
+        self._changed = True
         return self
 
     def unset(self, *doc_paths):
         """Removes attributes from documents in a collection.
 
         Args:
-            doc_path (string): The document path of the attribute to be
-                               removed.
+            doc_paths (list): The list of document paths of the attributes to be
+                              removed.
 
         Returns:
             mysqlx.ModifyStatement: ModifyStatement object.
         """
-        self._update_ops.extend([
-            UpdateSpec(mysqlxpb_enum(
+        for item in flexible_params(*doc_paths):
+            self._update_ops[item] = UpdateSpec(mysqlxpb_enum(
                 "Mysqlx.Crud.UpdateOperation.UpdateType.ITEM_REMOVE"), item)
-            for item in flexible_params(*doc_paths)])
+        self._changed = True
         return self
 
     def array_insert(self, field, value):
@@ -711,10 +831,10 @@ class ModifyStatement(FilterableStatement):
         Returns:
             mysqlx.ModifyStatement: ModifyStatement object.
         """
-        self._update_ops.append(
-            UpdateSpec(mysqlxpb_enum(
-                "Mysqlx.Crud.UpdateOperation.UpdateType.ARRAY_INSERT"),
-                       field, value))
+        self._update_ops[field] = UpdateSpec(mysqlxpb_enum(
+            "Mysqlx.Crud.UpdateOperation.UpdateType.ARRAY_INSERT"),
+                                             field, value)
+        self._changed = True
         return self
 
     def array_append(self, doc_path, value):
@@ -730,10 +850,10 @@ class ModifyStatement(FilterableStatement):
         Returns:
             mysqlx.ModifyStatement: ModifyStatement object.
         """
-        self._update_ops.append(
-            UpdateSpec(mysqlxpb_enum(
-                "Mysqlx.Crud.UpdateOperation.UpdateType.ARRAY_APPEND"),
-                       doc_path, value))
+        self._update_ops[doc_path] = UpdateSpec(mysqlxpb_enum(
+            "Mysqlx.Crud.UpdateOperation.UpdateType.ARRAY_APPEND"),
+                                                doc_path, value)
+        self._changed = True
         return self
 
     def patch(self, doc):
@@ -754,10 +874,10 @@ class ModifyStatement(FilterableStatement):
             raise ProgrammingError(
                 "Invalid data for update operation on document collection "
                 "table")
-        self._update_ops.append(
-            UpdateSpec(mysqlxpb_enum(
-                "Mysqlx.Crud.UpdateOperation.UpdateType.MERGE_PATCH"),
-                       '', doc.expr() if isinstance(doc, ExprParser) else doc))
+        self._update_ops["patch"] = UpdateSpec(
+            mysqlxpb_enum("Mysqlx.Crud.UpdateOperation.UpdateType.MERGE_PATCH"),
+            '', doc.expr() if isinstance(doc, ExprParser) else doc)
+        self._changed = True
         return self
 
     def execute(self):
@@ -771,7 +891,7 @@ class ModifyStatement(FilterableStatement):
         """
         if not self.has_where:
             raise ProgrammingError("No condition was found for modify")
-        return self._connection.update(self)
+        return self._connection.send_update(self)
 
 
 class ReadStatement(FilterableStatement):
@@ -885,7 +1005,7 @@ class ReadStatement(FilterableStatement):
         Returns:
             mysqlx.Result: Result object.
         """
-        return self._connection.find(self)
+        return self._connection.send_find(self)
 
 
 class FindStatement(ReadStatement):
@@ -936,7 +1056,6 @@ class SelectStatement(ReadStatement):
         super(SelectStatement, self).__init__(table, False)
         self._set_projection(*fields)
 
-
     def where(self, condition):
         """Sets the search condition to filter.
 
@@ -1026,7 +1145,7 @@ class UpdateStatement(FilterableStatement):
     """
     def __init__(self, table):
         super(UpdateStatement, self).__init__(target=table, doc_based=False)
-        self._update_ops = []
+        self._update_ops = {}
 
     def where(self, condition):
         """Sets the search condition to filter.
@@ -1068,9 +1187,9 @@ class UpdateStatement(FilterableStatement):
         Returns:
             mysqlx.UpdateStatement: UpdateStatement object.
         """
-        self._update_ops.append(
-            UpdateSpec(mysqlxpb_enum(
-                "Mysqlx.Crud.UpdateOperation.UpdateType.SET"), field, value))
+        self._update_ops[field] = UpdateSpec(mysqlxpb_enum(
+            "Mysqlx.Crud.UpdateOperation.UpdateType.SET"), field, value)
+        self._changed = True
         return self
 
     def execute(self):
@@ -1084,7 +1203,7 @@ class UpdateStatement(FilterableStatement):
         """
         if not self.has_where:
             raise ProgrammingError("No condition was found for update")
-        return self._connection.update(self)
+        return self._connection.send_update(self)
 
 
 class RemoveStatement(FilterableStatement):
@@ -1124,7 +1243,7 @@ class RemoveStatement(FilterableStatement):
         """
         if not self.has_where:
             raise ProgrammingError("No condition was found for remove")
-        return self._connection.delete(self)
+        return self._connection.send_delete(self)
 
 
 class DeleteStatement(FilterableStatement):
@@ -1172,7 +1291,7 @@ class DeleteStatement(FilterableStatement):
         """
         if not self.has_where:
             raise ProgrammingError("No condition was found for delete")
-        return self._connection.delete(self)
+        return self._connection.send_delete(self)
 
 
 class CreateCollectionIndexStatement(Statement):
@@ -1234,12 +1353,12 @@ class CreateCollectionIndexStatement(Statement):
 
         # Validate members that constraint the index
         if not self._fields_desc:
-            raise ProgrammingError("Required member \"fields\" not found in "
+            raise ProgrammingError("Required member 'fields' not found in "
                                    "the given index description: {}"
                                    "".format(self._index_desc))
 
         if not isinstance(self._fields_desc, list):
-            raise ProgrammingError("Required member \"fields\" must contain a "
+            raise ProgrammingError("Required member 'fields' must contain a "
                                    "list.")
 
         args = {}
@@ -1266,39 +1385,42 @@ class CreateCollectionIndexStatement(Statement):
                 constraint["member"] = field_desc.pop("field")
                 constraint["type"] = field_desc.pop("type")
                 constraint["required"] = field_desc.pop("required", False)
+                constraint["array"] = field_desc.pop("array", False)
+                if not isinstance(constraint["required"], bool):
+                    raise TypeError("Field member 'required' must be Boolean")
+                if not isinstance(constraint["array"], bool):
+                    raise TypeError("Field member 'array' must be Boolean")
                 if args["type"].upper() == "SPATIAL" and \
                    not constraint["required"]:
-                    raise ProgrammingError('Field member "required" must be '
-                                           'set to "True" when index type is'
-                                           ' set to "SPATIAL"')
+                    raise ProgrammingError(
+                        "Field member 'required' must be set to 'True' when "
+                        "index type is set to 'SPATIAL'")
                 if args["type"].upper() == "INDEX" and \
-                   constraint["type"] == 'GEOJSON':
-                    raise ProgrammingError('Index "type" must be set to '
-                                           '"SPATIAL" when field type is set '
-                                           'to "GEOJSON"')
+                   constraint["type"] == "GEOJSON":
+                    raise ProgrammingError(
+                        "Index 'type' must be set to 'SPATIAL' when field "
+                        "type is set to 'GEOJSON'")
                 if "collation" in field_desc:
                     if not constraint["type"].upper().startswith("TEXT"):
                         raise ProgrammingError(
-                            "The \"collation\" member can only be used when "
-                            "field  type is set to \"GEOJSON\"")
-                    else:
-                        constraint["collation"] = field_desc.pop("collation")
+                            "The 'collation' member can only be used when "
+                            "field type is set to '{}'"
+                            "".format(constraint["type"].upper()))
+                    constraint["collation"] = field_desc.pop("collation")
                 # "options" and "srid" fields in IndexField can be
                 # present only if "type" is set to "GEOJSON"
                 if "options" in field_desc:
-                    if constraint["type"].upper() != 'GEOJSON':
+                    if constraint["type"].upper() != "GEOJSON":
                         raise ProgrammingError(
-                            "The \"options\" member can only be used when "
-                            "index type is set to \"GEOJSON\"")
-                    else:
-                        constraint["options"] = field_desc.pop("options")
+                            "The 'options' member can only be used when "
+                            "index type is set to 'GEOJSON'")
+                    constraint["options"] = field_desc.pop("options")
                 if "srid" in field_desc:
-                    if constraint["type"].upper() != 'GEOJSON':
+                    if constraint["type"].upper() != "GEOJSON":
                         raise ProgrammingError(
-                            "The \"srid\" member can only be used when index"
-                            " type is set to \"GEOJSON\"")
-                    else:
-                        constraint["srid"] = field_desc.pop("srid")
+                            "The 'srid' member can only be used when index "
+                            "type is set to 'GEOJSON'")
+                    constraint["srid"] = field_desc.pop("srid")
                 args["constraint"].append(constraint)
         except KeyError as err:
             raise ProgrammingError("Required inner member {} not found in "
